/* Generated by wayland-scanner 1.23.1 */

/*
 * SPDX-FileCopyrightText: 2015 Martin Gräßlin
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include "wayland-util.h"

#ifndef __has_attribute
# define __has_attribute(x) 0  /* Compatibility with non-clang compilers. */
#endif

#if (__has_attribute(visibility) || defined(__GNUC__) && __GNUC__ >= 4)
#define WL_PRIVATE __attribute__ ((visibility("hidden")))
#else
#define WL_PRIVATE
#endif

extern const struct wl_interface org_kde_kwin_dpms_interface;
extern const struct wl_interface wl_output_interface;

static const struct wl_interface *dpms_types[] = {
	NULL,
	&org_kde_kwin_dpms_interface,
	&wl_output_interface,
};

static const struct wl_message org_kde_kwin_dpms_manager_requests[] = {
	{ "get", "no", dpms_types + 1 },
};

WL_PRIVATE const struct wl_interface org_kde_kwin_dpms_manager_interface = {
	"org_kde_kwin_dpms_manager", 1,
	1, org_kde_kwin_dpms_manager_requests,
	0, NULL,
};

static const struct wl_message org_kde_kwin_dpms_requests[] = {
	{ "set", "u", dpms_types + 0 },
	{ "release", "", dpms_types + 0 },
};

static const struct wl_message org_kde_kwin_dpms_events[] = {
	{ "supported", "u", dpms_types + 0 },
	{ "mode", "u", dpms_types + 0 },
	{ "done", "", dpms_types + 0 },
};

WL_PRIVATE const struct wl_interface org_kde_kwin_dpms_interface = {
	"org_kde_kwin_dpms", 1,
	2, org_kde_kwin_dpms_requests,
	3, org_kde_kwin_dpms_events,
};

