"#error Do not run app-defaults files through xrdb!",
"#error That does not do what you might expect.",
"#error Put this file in /usr/lib/X11/app-defaults/XScreenSaver instead.",
"*mode:			random",
"*timeout:		0:10:00",
"*cycle:			0:10:00",
"*lockTimeout:		0:00:00",
"*passwdTimeout:		0:00:30",
"*dpmsEnabled:		False",
"*dpmsQuickoffEnabled:	False",
"*dpmsStandby:		2:00:00",
"*dpmsSuspend:		2:00:00",
"*dpmsOff:		4:00:00",
"*grabDesktopImages:	True",
"*grabVideoFrames:	False",
"*chooseRandomImages:	True",
"*imageDirectory:	/Library/Desktop Pictures/",
"*nice:			10",
"*memoryLimit:		0",
"*lock:			False",
"*verbose:		False",
"*fade:			True",
"*unfade:		True",
"*fadeSeconds:		0:00:03",
"*fadeTicks:		20",
"*splash:		True",
"*splashDuration:	0:00:05",
"*visualID:		default",
"*installColormap:	True",
"*ignoreUninstalledPrograms: False",
"*authWarningSlack:	20",
"*textMode:		url",
"*textLiteral:		XScreenSaver",
"*textFile:		",
"*textProgram:		fortune",
"*textURL:		https://en.wikipedia.org/w/index.php?title=Special:NewPages&feed=rss",
"*demoCommand: xscreensaver-settings",
"*helpURL: https://www.jwz.org/xscreensaver/man.html",
"*loadURL: x-www-browser '%s' || firefox '%s' || chromium-browser '%s'",
"*manualCommand: xdg-open man:%s || \
		    x-terminal-emulator -e /bin/sh -c 'man \"%s\"'",
"*dateFormat:		%I:%M %p, %a %b %e",
"*newLoginCommand:	no-such-login-manager",
"XScreenSaver.pointerHysteresis:		10",
"XScreenSaver.bourneShell:		/bin/sh",
"*dialogTheme:			default",
"*themeNames: Default, Borderless, Dark Gray, Borderless Black, \
             Green Black, White, Blue, Aqua Black, Wine",
"*Dialog.headingFont:	sans-serif bold 16",
"*Dialog.bodyFont:	sans-serif 14",
"*Dialog.errorFont:	sans-serif bold 14",
"*Dialog.labelFont:	sans-serif bold 14",
"*Dialog.unameFont:	sans-serif 12",
"*Dialog.buttonFont:	sans-serif bold 14",
"*Dialog.dateFont:	sans-serif 9",
"*passwd.uname:		True",
"*passwd.asterisks:	True",
"*default.Dialog.foreground:			#000000",
"*default.Dialog.background:			#E6E6E6",
"*default.Dialog.button.foreground:		#000000",
"*default.Dialog.button.background:		#F5F5F5",
"*default.Dialog.logo.background:		#BBBBBB",
"*default.Dialog.text.foreground:		#000000",
"*default.Dialog.error.foreground:		#FF0000",
"*default.Dialog.text.background:		#FFFFFF",
"*default.Dialog.topShadowColor:			#FFFFFF",
"*default.Dialog.bottomShadowColor:		#CECECE",
"*default.Dialog.shadowWidth:			2",
"*default.Dialog.logo.width:			210",
"*default.Dialog.logo.height:			210",
"*default.Dialog.thermometer.foreground:		#4464AC",
"*default.Dialog.thermometer.background:		#FFFFFF",
"*default.Dialog.thermometer.width:		8",
"*default.Dialog.borderColor:			#CECECE",
"*default.Dialog.borderWidth:			0",
"*default.Dialog.internalPadding:		24",
"*borderless.Dialog.topShadowColor:		#E6E6E6",
"*borderless.Dialog.button.background:		#FFFFFF",
"*borderless.Dialog.bottomShadowColor:		#E6E6E6",
"*borderless.Dialog.logo.background:		#E6E6E6",
"*borderless.Dialog.borderColor:			#888888",
"*borderless.Dialog.thermometer.width:		6",
"*borderless.Dialog.borderWidth:			1",
"*darkgray.Dialog.foreground:			#CCCCCC",
"*darkgray.Dialog.background:			#333333",
"*darkgray.Dialog.topShadowColor:		#444444",
"*darkgray.Dialog.bottomShadowColor:		#111111",
"*darkgray.Dialog.borderColor:			#111111",
"*darkgray.Dialog.text.foreground:		#DDDDDD",
"*darkgray.Dialog.text.background:		#666666",
"*darkgray.Dialog.button.foreground:		#CCCCCC",
"*darkgray.Dialog.button.background:		#666666",
"*darkgray.Dialog.logo.background:		#444444",
"*darkgray.Dialog.thermometer.foreground:	#4464AC",
"*darkgray.Dialog.thermometer.background:	#666666",
"*darkgray.Dialog.borderWidth:			0",
"*borderlessblack.Dialog.foreground:		#CCCCCC",
"*borderlessblack.Dialog.background:		#000000",
"*borderlessblack.Dialog.topShadowColor:		#000000",
"*borderlessblack.Dialog.bottomShadowColor:	#000000",
"*borderlessblack.Dialog.text.foreground:	#CCCCCC",
"*borderlessblack.Dialog.text.background:	#000000",
"*borderlessblack.Dialog.button.foreground:	#CCCCCC",
"*borderlessblack.Dialog.button.background:	#333333",
"*borderlessblack.Dialog.logo.background:	#000000",
"*borderlessblack.Dialog.thermometer.foreground:	#CCCCCC",
"*borderlessblack.Dialog.thermometer.background:	#333333",
"*borderlessblack.Dialog.thermometer.width:	3",
"*borderlessblack.Dialog.borderColor:		#333333",
"*borderlessblack.Dialog.borderWidth:		1",
"*greenblack.Dialog.foreground:			#00FF00",
"*greenblack.Dialog.background:			#000000",
"*greenblack.Dialog.topShadowColor:		#000000",
"*greenblack.Dialog.bottomShadowColor:		#000000",
"*greenblack.Dialog.shadowWidth:			1",
"*greenblack.Dialog.text.foreground:		#00FF00",
"*greenblack.Dialog.text.background:		#006600",
"*greenblack.Dialog.button.foreground:		#00FF00",
"*greenblack.Dialog.button.background:		#006600",
"*greenblack.Dialog.logo.background:		#000000",
"*greenblack.Dialog.thermometer.foreground:	#00CC00",
"*greenblack.Dialog.thermometer.background:	#006600",
"*greenblack.Dialog.borderColor:			#006600",
"*greenblack.Dialog.borderWidth:			1",
"*white.Dialog.foreground:			#000000",
"*white.Dialog.background:			#FFFFFF",
"*white.Dialog.topShadowColor:			#CCCCCC",
"*white.Dialog.bottomShadowColor:		#CCCCCC",
"*white.Dialog.shadowWidth:			1",
"*white.Dialog.borderColor:			#CCCCCC",
"*white.Dialog.text.foreground:			#000000",
"*white.Dialog.text.background:			#FFFFFF",
"*white.Dialog.button.foreground:		#000000",
"*white.Dialog.button.background:		#FFFFFF",
"*white.Dialog.logo.background:			#FFFFFF",
"*white.Dialog.borderWidth:			0",
"*blue.Dialog.foreground:			#000000",
"*blue.Dialog.background:			#BBCCDD",
"*blue.Dialog.topShadowColor:			#CCDDEE",
"*blue.Dialog.bottomShadowColor:			#AABBCC",
"*blue.Dialog.borderColor:			#AABBCC",
"*blue.Dialog.text.foreground:			#000000",
"*blue.Dialog.text.background:			#DDEEFF",
"*blue.Dialog.button.foreground:			#000000",
"*blue.Dialog.button.background:			#DDEEFF",
"*blue.Dialog.logo.background:			#BBCCDD",
"*blue.Dialog.thermometer.foreground:		#5566AA",
"*blue.Dialog.thermometer.background:		#BBCCDD",
"*blue.Dialog.borderWidth:			0",
"*aquablack.Dialog.foreground:			#00EFEF",
"*aquablack.Dialog.background:			#000000",
"*aquablack.Dialog.topShadowColor:		#000000",
"*aquablack.Dialog.bottomShadowColor:		#000000",
"*aquablack.Dialog.shadowWidth:			2",
"*aquablack.Dialog.button.foreground:		#000000",
"*aquablack.Dialog.button.background:		#2244EE",
"*aquablack.Dialog.logo.background:		#000000",
"*aquablack.Dialog.text.foreground:		#2244EE",
"*aquablack.Dialog.text.background:		#EEEEEE",
"*aquablack.Dialog.thermometer.foreground:	#2244EE",
"*aquablack.Dialog.thermometer.background:	#000088",
"*aquablack.Dialog.borderColor:			#000066",
"*aquablack.Dialog.borderWidth:			1",
"*aquablack.Dialog.internalPadding:		36",
"*wine.Dialog.foreground:			#AD8FA6",
"*wine.Dialog.background:			#2C041E",
"*wine.Dialog.topShadowColor:			#2C041E",
"*wine.Dialog.bottomShadowColor:			#2C041E",
"*wine.Dialog.text.foreground:			#706B70",
"*wine.Dialog.text.background:			#F9F9F8",
"*wine.Dialog.button.foreground:			#5F585B",
"*wine.Dialog.logo.background:			#2C041E",
"*wine.Dialog.thermometer.foreground:		#AD8FA6",
"*wine.Dialog.thermometer.background:		#4D2946",
"*wine.Dialog.borderColor:			#4D2946",
"*wine.Dialog.thermometer.width:			6",
"*wine.Dialog.borderWidth:			1",
"*errorFont:  sans-serif bold 18",
"*errorColor: #FF0000",
"*programs:								      \
				maze --root				    \\n\
  GL: 				superquadrics --root			    \\n\
				attraction --root			    \\n\
				blitspin --root				    \\n\
				greynetic --root			    \\n\
				helix --root				    \\n\
				hopalong --root				    \\n\
				imsmap --root				    \\n\
-				noseguy --root				    \\n\
-				pyro --root				    \\n\
				qix --root				    \\n\
-				rocks --root				    \\n\
				rorschach --root			    \\n\
				decayscreen --root			    \\n\
				flame --root				    \\n\
				halo --root				    \\n\
				slidescreen --root			    \\n\
				pedal --root				    \\n\
				bouboule --root				    \\n\
-				braid --root				    \\n\
				coral --root				    \\n\
				deco --root				    \\n\
				drift --root				    \\n\
-				fadeplot --root				    \\n\
				galaxy --root				    \\n\
				goop --root				    \\n\
				grav --root				    \\n\
				ifs --root				    \\n\
  GL: 				jigsaw --root				    \\n\
				julia --root				    \\n\
-				kaleidescope --root			    \\n\
  GL: 				moebius --root				    \\n\
				moire --root				    \\n\
  GL: 				morph3d --root				    \\n\
				mountain --root				    \\n\
				munch --root				    \\n\
				penrose --root				    \\n\
  GL: 				pipes --root				    \\n\
				rdbomb --root				    \\n\
  GL: 				rubik --root				    \\n\
-				sierpinski --root			    \\n\
				slip --root				    \\n\
  GL: 				sproingies --root			    \\n\
				starfish --root				    \\n\
				strange --root				    \\n\
				swirl --root				    \\n\
				triangle --root				    \\n\
				xjack --root				    \\n\
				xlyap --root				    \\n\
  GL: 				atlantis --root				    \\n\
				bsod --root				    \\n\
  GL: 				bubble3d --root				    \\n\
  GL: 				cage --root				    \\n\
-				crystal --root				    \\n\
				cynosure --root				    \\n\
				discrete --root				    \\n\
				distort --root				    \\n\
				epicycle --root				    \\n\
				flow --root				    \\n\
  GL: 				glplanet --root				    \\n\
				interference --root			    \\n\
				kumppa --root				    \\n\
  GL: 				lament --root				    \\n\
				moire2 --root				    \\n\
  GL: 				sonar --root				    \\n\
  GL: 				stairs --root				    \\n\
				truchet --root				    \\n\
-				vidwhacker --root			    \\n\
-				webcollage --root			    \\n\
				blaster --root				    \\n\
				bumps --root				    \\n\
				ccurve --root				    \\n\
				compass --root				    \\n\
				deluxe --root				    \\n\
-				demon --root				    \\n\
- GL: 				extrusion --root			    \\n\
-				loop --root				    \\n\
				penetrate --root			    \\n\
				petri --root				    \\n\
				phosphor --root				    \\n\
  GL: 				pulsar --root				    \\n\
				ripples --root				    \\n\
				shadebobs --root			    \\n\
  GL: 				sierpinski3d --root			    \\n\
				spotlight --root			    \\n\
				squiral --root				    \\n\
				wander --root				    \\n\
				xflame --root				    \\n\
				xmatrix --root				    \\n\
  GL: 				gflux --root				    \\n\
-				nerverot --root				    \\n\
				xrayswarm --root			    \\n\
				xspirograph --root			    \\n\
  GL: 				circuit --root				    \\n\
  GL: 				dangerball --root			    \\n\
- GL: 				dnalogo --root				    \\n\
  GL: 				engine --root				    \\n\
  GL: 				flipscreen3d --root			    \\n\
  GL: 				gltext --root				    \\n\
  GL: 				menger --root				    \\n\
  GL: 				molecule --root				    \\n\
				rotzoomer --root			    \\n\
				scooter --root				    \\n\
				speedmine --root			    \\n\
  GL: 				starwars --root				    \\n\
  GL: 				stonerview --root			    \\n\
				vermiculate --root			    \\n\
				whirlwindwarp --root			    \\n\
				zoom --root				    \\n\
				anemone --root				    \\n\
				apollonian --root			    \\n\
  GL: 				boxed --root				    \\n\
  GL: 				cubenetic --root			    \\n\
  GL: 				endgame --root				    \\n\
				euler2d --root				    \\n\
				fluidballs --root			    \\n\
  GL: 				flurry --root				    \\n\
- GL: 				glblur --root				    \\n\
  GL: 				glsnake --root				    \\n\
				halftone --root				    \\n\
  GL: 				juggler3d --root			    \\n\
  GL: 				lavalite --root				    \\n\
-				polyominoes --root			    \\n\
  GL: 				queens --root				    \\n\
- GL: 				sballs --root				    \\n\
  GL: 				spheremonics --root			    \\n\
				twang --root				    \\n\
- GL: 				antspotlight --root			    \\n\
				apple2 --root				    \\n\
  GL: 				atunnel --root				    \\n\
				barcode --root				    \\n\
  GL: 				blinkbox --root				    \\n\
  GL: 				blocktube --root			    \\n\
  GL: 				bouncingcow --root			    \\n\
				cloudlife --root			    \\n\
  GL: 				cubestorm --root			    \\n\
				eruption --root				    \\n\
  GL: 				flipflop --root				    \\n\
  GL: 				flyingtoasters --root			    \\n\
				fontglide --root			    \\n\
  GL: 				gleidescope --root			    \\n\
  GL: 				glknots --root				    \\n\
  GL: 				glmatrix --root				    \\n\
- GL: 				glslideshow --root			    \\n\
  GL: 				hypertorus --root			    \\n\
- GL: 				jigglypuff --root			    \\n\
				metaballs --root			    \\n\
  GL: 				mirrorblob --root			    \\n\
				piecewise --root			    \\n\
  GL: 				polytopes --root			    \\n\
				pong --root				    \\n\
				popsquares --root			    \\n\
  GL: 				surfaces --root				    \\n\
				xanalogtv --root			    \\n\
				abstractile --root			    \\n\
				anemotaxis --root			    \\n\
- GL: 				antinspect --root			    \\n\
				fireworkx --root			    \\n\
				fuzzyflakes --root			    \\n\
				interaggregate --root			    \\n\
				intermomentary --root			    \\n\
				memscroller --root			    \\n\
  GL: 				noof --root				    \\n\
				pacman --root				    \\n\
  GL: 				pinion --root				    \\n\
  GL: 				polyhedra --root			    \\n\
- GL: 				providence --root			    \\n\
				substrate --root			    \\n\
				wormhole --root				    \\n\
- GL: 				antmaze --root				    \\n\
  GL: 				boing --root				    \\n\
				boxfit --root				    \\n\
  GL: 				carousel --root				    \\n\
				celtic --root				    \\n\
  GL: 				crackberg --root			    \\n\
  GL: 				cube21 --root				    \\n\
				fiberlamp --root			    \\n\
  GL: 				fliptext --root				    \\n\
  GL: 				glhanoi --root				    \\n\
  GL: 				tangram --root				    \\n\
  GL: 				timetunnel --root			    \\n\
  GL: 				glschool --root				    \\n\
  GL: 				topblock --root				    \\n\
  GL: 				cubicgrid --root			    \\n\
				cwaves --root				    \\n\
  GL: 				gears --root				    \\n\
  GL: 				glcells --root				    \\n\
  GL: 				lockward --root				    \\n\
				m6502 --root				    \\n\
  GL: 				moebiusgears --root			    \\n\
  GL: 				voronoi --root				    \\n\
  GL: 				hypnowheel --root			    \\n\
  GL: 				klein --root				    \\n\
-				lcdscrub --root				    \\n\
  GL: 				photopile --root			    \\n\
  GL: 				skytentacles --root			    \\n\
  GL: 				rubikblocks --root			    \\n\
  GL: 				companioncube --root			    \\n\
  GL: 				hilbert --root				    \\n\
  GL: 				tronbit --root				    \\n\
  GL: 				geodesic --root				    \\n\
				hexadrop --root				    \\n\
  GL: 				kaleidocycle --root			    \\n\
  GL: 				quasicrystal --root			    \\n\
  GL: 				starnest --root				    \\n\
  GL: 				unknownpleasures --root			    \\n\
  GL: 				alienbeacon --root			    \\n\
				binaryring --root			    \\n\
  GL: 				cityflow --root				    \\n\
  GL: 				geodesicgears --root			    \\n\
  GL: 				projectiveplane --root			    \\n\
  GL: 				romanboy --root				    \\n\
				tessellimage --root			    \\n\
  GL: 				topologica --root			    \\n\
  GL: 				winduprobot --root			    \\n\
  GL: 				fluxcore --root				    \\n\
  GL: 				gimbalharmonics --root			    \\n\
  GL: 				protophore --root			    \\n\
  GL: 				skyline --root				    \\n\
  GL: 				splitflap --root			    \\n\
  GL: 				cubestack --root			    \\n\
  GL: 				cubetwist --root			    \\n\
  GL: 				discoball --root			    \\n\
  GL: 				dymaxionmap --root			    \\n\
  GL: 				elementalring --root			    \\n\
  GL: 				energystream --root			    \\n\
  GL: 				hexstrut --root				    \\n\
  GL: 				hydrostat --root			    \\n\
  GL: 				raverhoop --root			    \\n\
  GL: 				splodesic --root			    \\n\
  GL: 				unicrud --root				    \\n\
  GL: 				esper --root				    \\n\
  GL: 				vigilance --root			    \\n\
  GL: 				crumbler --root				    \\n\
				filmleader --root			    \\n\
				glitchpeg --root			    \\n\
  GL: 				handsy --root				    \\n\
  GL: 				maze3d --root				    \\n\
  GL: 				peepers --root				    \\n\
  GL: 				razzledazzle --root			    \\n\
				vfeedback --root			    \\n\
  GL: 				deepstars --root			    \\n\
  GL: 				gravitywell --root			    \\n\
  GL: 				synthwavecity --root			    \\n\
  GL: 				batteredplanet --root			    \\n\
  GL: 				beats --root				    \\n\
  GL: 				covid19 --root				    \\n\
  GL: 				etruscanvenus --root			    \\n\
  GL: 				gibson --root				    \\n\
  GL: 				headroom --root				    \\n\
  GL: 				sphereeversion --root			    \\n\
				binaryhorizon --root			    \\n\
				marbling --root				    \\n\
  GL: 				neontriangulator --root			    \\n\
  GL: 				stripeytorus --root			    \\n\
  GL: 				chompytower --root			    \\n\
  GL: 				hextrail --root				    \\n\
  GL: 				mapscroller --root			    \\n\
  GL: 				nakagin --root				    \\n\
  GL: 				squirtorus --root			    \\n\
  GL: 				stardome --root				    \\n\
  GL: 				cubocteversion --root			    \\n\
				droste --root				    \\n\
  GL: 				logarithmiccircles --root		    \\n\
  GL: 				neongravity --root			    \\n\
  GL: 				papercube --root			    \\n\
  GL: 				skulloop --root				    \\n\
  GL: 				highvoltage --root			    \\n\
  GL: 				kallisti --root				    \\n\
  GL: 				klondike --root				    \\n\
  GL: 				selfreflect --root			    \\n\
  GL: 				truchetzoom --root			    \\n\
  GL: 				dumpsterfire --root			    \\n\
  GL: 				hexplasma --root			    \\n\
  GL: 				hopffibration --root			    \\n\
  GL: 				platonicfolding --root			    \\n",
"*hacks.alienbeacon.name:    Alien Beacon",
"*hacks.antinspect.name:     Ant Inspect",
"*hacks.antmaze.name:        Ant Maze",
"*hacks.antspotlight.name:   Ant Spotlight",
"*hacks.apple2.name:         Apple ][",
"*hacks.batteredplanet.name: Battered Planet",
"*hacks.binaryhorizon.name:  Binary Horizon",
"*hacks.binaryring.name:     Binary Ring",
"*hacks.blinkbox.name:       Blink Box",
"*hacks.blitspin.name:       Blit Spin",
"*hacks.blocktube.name:      Block Tube",
"*hacks.bouncingcow.name:    Bouncing Cow",
"*hacks.boxfit.name:         Box Fit",
"*hacks.bsod.name:           BSOD",
"*hacks.bubble3d.name:       Bubble 3D",
"*hacks.ccurve.name:         C Curve",
"*hacks.chompytower.name:    Chompy Tower",
"*hacks.cityflow.name:       City Flow",
"*hacks.cloudlife.name:      Cloud Life",
"*hacks.companioncube.name:  Companion Cube",
"*hacks.covid19.name:        COVID19",
"*hacks.cube21.name:         Cube 21",
"*hacks.cubestack.name:      Cube Stack",
"*hacks.cubestorm.name:      Cube Storm",
"*hacks.cubetwist.name:      Cube Twist",
"*hacks.cubicgrid.name:      Cubic Grid",
"*hacks.cubocteversion.name: Cuboctahedron Eversion",
"*hacks.cwaves.name:         C Waves",
"*hacks.dangerball.name:     Danger Ball",
"*hacks.decayscreen.name:    Decay Screen",
"*hacks.deepstars.name:      Deep Stars",
"*hacks.dnalogo.name:        DNA Logo",
"*hacks.dumpsterfire.name:   Dumpster Fire",
"*hacks.dymaxionmap.name:    Dymaxion Map",
"*hacks.elementalring.name:  Elemental Ring",
"*hacks.energystream.name:   Energy Stream",
"*hacks.etruscanvenus.name:  Etruscan Venus",
"*hacks.euler2d.name:        Euler 2D",
"*hacks.fadeplot.name:       Fade Plot",
"*hacks.fiberlamp.name:      Fiber Lamp",
"*hacks.filmleader.name:     Film Leader",
"*hacks.flipflop.name:       Flip Flop",
"*hacks.flipscreen3d.name:   Flip Screen 3D",
"*hacks.fliptext.name:       Flip Text",
"*hacks.fluidballs.name:     Fluid Balls",
"*hacks.fluxcore.name:       Flux Core",
"*hacks.flyingtoasters.name: Flying Toasters",
"*hacks.fontglide.name:      Font Glide",
"*hacks.fuzzyflakes.name:    Fuzzy Flakes",
"*hacks.geodesicgears.name:  Geodesic Gears",
"*hacks.gflux.name:          GFlux",
"*hacks.gimbalharmonics.name:Gimbal Harmonics",
"*hacks.glblur.name:         GL Blur",
"*hacks.glcells.name:        GL Cells",
"*hacks.gleidescope.name:    Gleidescope",
"*hacks.glforestfire.name:   GL Forest Fire",
"*hacks.glhanoi.name:        GL Hanoi",
"*hacks.glitchpeg.name:      GlitchPEG",
"*hacks.glknots.name:        GL Knots",
"*hacks.glmatrix.name:       GL Matrix",
"*hacks.glplanet.name:       GL Planet",
"*hacks.glschool.name:       GL School",
"*hacks.glslideshow.name:    GL Slideshow",
"*hacks.glsnake.name:        GL Snake",
"*hacks.gltext.name:         GL Text",
"*hacks.gravitywell.name:    Gravity Well",
"*hacks.hexplasma.name:      Hex Plasma",
"*hacks.hexstrut.name:       Hex Strut",
"*hacks.hextrail.name:       Hex Trail",
"*hacks.highvoltage.name:    High Voltage",
"*hacks.hopffibration.name:  Hopf Fibration",
"*hacks.ifs.name:            IFS",
"*hacks.imsmap.name:         IMS Map",
"*hacks.jigglypuff.name:     Jiggly Puff",
"*hacks.juggler3d.name:      Juggler 3D",
"*hacks.lcdscrub.name:       LCD Scrub",
"*hacks.lmorph.name:         LMorph",
"*hacks.logarithmiccircles.name:Logarithmic Circles",
"*hacks.m6502.name:          m6502",
"*hacks.mapscroller.name:    Map Scroller",
"*hacks.maze3d.name:         Maze 3D",
"*hacks.memscroller.name:    Mem Scroller",
"*hacks.metaballs.name:      Meta Balls",
"*hacks.mirrorblob.name:     Mirror Blob",
"*hacks.moebius.name:        Möbius",
"*hacks.moebiusgears.name:   Möbius Gears",
"*hacks.moire.name:          Moiré",
"*hacks.moire2.name:         Moiré 2",
"*hacks.morph3d.name:        Morph 3D",
"*hacks.neongravity.name:    Neon Gravity",
"*hacks.neontriangulator.name:Neon Triangulator",
"*hacks.nerverot.name:       Nerve Rot",
"*hacks.noseguy.name:        Nose Guy",
"*hacks.pacman.name:         Pac-Man",
"*hacks.papercube.name:      Paper Cube",
"*hacks.photopile.name:      Photo Pile",
"*hacks.platonicfolding.name:Platonic Folding",
"*hacks.popsquares.name:     Pop Squares",
"*hacks.projectiveplane.name:Projective Plane",
"*hacks.quasicrystal.name:   Quasi-Crystal",
"*hacks.raverhoop.name:      Raver Hoop",
"*hacks.razzledazzle.name:   Razzle Dazzle",
"*hacks.rdbomb.name:         RD-Bomb",
"*hacks.romanboy.name:       Roman Boy",
"*hacks.rotzoomer.name:      Rot Zoomer",
"*hacks.rubikblocks.name:    Rubik Blocks",
"*hacks.sballs.name:         SBalls",
"*hacks.selfreflect.name:    Self Reflect",
"*hacks.shadebobs.name:      Shade Bobs",
"*hacks.sierpinski3d.name:   Sierpinski 3D",
"*hacks.skytentacles.name:   Sky Tentacles",
"*hacks.slidescreen.name:    Slide Screen",
"*hacks.speedmine.name:      Speed Mine",
"*hacks.sphereeversion.name: Sphere Eversion",
"*hacks.splitflap.name:      Split-Flap",
"*hacks.stardome.name:       Star Dome",
"*hacks.starnest.name:       Star Nest",
"*hacks.starwars.name:       Star Wars",
"*hacks.stonerview.name:     Stoner View",
"*hacks.stripeytorus.name:   Stripey Torus",
"*hacks.synthwavecity.name:  Synthwave City",
"*hacks.t3d.name:            T3D",
"*hacks.testx11.name:        Test X11",
"*hacks.timetunnel.name:     Time Tunnel",
"*hacks.topblock.name:       Top Block",
"*hacks.tronbit.name:        Tron Bit",
"*hacks.truchetzoom.name:    Truchet Zoom",
"*hacks.unknownpleasures.name:Unknown Pleasures",
"*hacks.vfeedback.name:      VFeedback",
"*hacks.vidwhacker.name:     Vid Whacker",
"*hacks.webcollage.name:     Web Collage",
"*hacks.whirlwindwarp.name:  Whirlwind Warp",
"*hacks.winduprobot.name:    Windup Robot",
"*hacks.xanalogtv.name:      XAnalogTV",
"*hacks.xrayswarm.name:      XRaySwarm",
