/* Copyright © 2025 Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

#ifndef __DOUBLETIME_H__
#define __DOUBLETIME_H__

/* Current time as a double, with (probably) microsecond accuracy. */
extern double double_time (void);

#endif /* __DOUBLETIME_H__ */
