/* Generated by wayland-scanner 1.23.1 */

/*
 * Copyright (C) 2015 Martin Gräßlin
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdint.h>
#include "wayland-util.h"

#ifndef __has_attribute
# define __has_attribute(x) 0  /* Compatibility with non-clang compilers. */
#endif

#if (__has_attribute(visibility) || defined(__GNUC__) && __GNUC__ >= 4)
#define WL_PRIVATE __attribute__ ((visibility("hidden")))
#else
#define WL_PRIVATE
#endif

extern const struct wl_interface org_kde_kwin_idle_timeout_interface;
extern const struct wl_interface wl_seat_interface;

static const struct wl_interface *idle_types[] = {
	&org_kde_kwin_idle_timeout_interface,
	&wl_seat_interface,
	NULL,
};

static const struct wl_message org_kde_kwin_idle_requests[] = {
	{ "get_idle_timeout", "nou", idle_types + 0 },
};

WL_PRIVATE const struct wl_interface org_kde_kwin_idle_interface = {
	"org_kde_kwin_idle", 1,
	1, org_kde_kwin_idle_requests,
	0, NULL,
};

static const struct wl_message org_kde_kwin_idle_timeout_requests[] = {
	{ "release", "", idle_types + 0 },
	{ "simulate_user_activity", "", idle_types + 0 },
};

static const struct wl_message org_kde_kwin_idle_timeout_events[] = {
	{ "idle", "", idle_types + 0 },
	{ "resumed", "", idle_types + 0 },
};

WL_PRIVATE const struct wl_interface org_kde_kwin_idle_timeout_interface = {
	"org_kde_kwin_idle_timeout", 1,
	2, org_kde_kwin_idle_timeout_requests,
	2, org_kde_kwin_idle_timeout_events,
};

