"*fontList:                       *-helvetica-medium-r-*-*-*-120-*-*-*-iso8859-1",
"*demoDialog*label1.fontList:     *-helvetica-medium-r-*-*-*-140-*-*-*-iso8859-1",
"*cmdText.fontList:                 *-courier-medium-r-*-*-*-120-*-*-*-iso8859-1",
"*label0.fontList:                  *-helvetica-bold-r-*-*-*-140-*-*-*-iso8859-1",
"XScreenSaver*doc.fontList:       *-helvetica-medium-r-*-*-*-100-*-*-*-iso8859-1",
"*foreground:			#000000",
"*background:			#C0C0C0",
"*XmTextField.foreground:	#000000",
"*XmTextField.background:	#FFFFFF",
"*list.foreground:		#000000",
"*list.background:		#FFFFFF",
"*ApplicationShell.title:	XScreenSaver",
"*warning.title:			XScreenSaver",
"*warning_popup.title:		XScreenSaver",
"*allowShellResize:		True",
"*autoUnmanage:			False",
"*menubar*file.labelString:	File",
"*menubar*file.mnemonic:		F",
"*file.blank.labelString:	Blank Screen Now",
"*file.blank.mnemonic:		B",
"*file.lock.labelString:		Lock Screen Now",
"*file.lock.mnemonic:		L",
"*file.kill.labelString:		Kill Daemon",
"*file.kill.mnemonic:		K",
"*file.restart.labelString:	Restart Daemon",
"*file.restart.mnemonic:		R",
"*file.exit.labelString:		Exit",
"*file.exit.mnemonic:		E",
"*menubar*edit.labelString:	Edit",
"*menubar*edit.mnemonic:		E",
"*edit.cut.labelString:		Cut",
"*edit.cut.mnemonic:		u",
"*edit.copy.labelString:		Copy",
"*edit.copy.mnemonic:		C",
"*edit.paste.labelString:	Paste",
"*edit.paste.mnemonic:		P",
"*menubar*help.labelString:	Help",
"*menubar*help.mnemonic:		H",
"*help.about.labelString:	About...",
"*help.about.mnemonic:		A",
"*help.docMenu.labelString:	Documentation...",
"*help.docMenu.mnemonic:		D",
"*demoTab.marginWidth:		10",
"*optionsTab.marginWidth:	10",
"*XmScrolledWindow.topOffset:	10",
"*XmScrolledWindow.leftOffset:	10",
"*demoTab.topOffset:		4",
"*form1.bottomOffset:		10",
"*form3.leftOffset:		10",
"*form3.rightOffset:		10",
"*frame.topOffset:		10",
"*frame.bottomOffset:		10",
"*enabled.topOffset:		10",
"*visLabel.topOffset:		10",
"*combo.topOffset:		10",
"*form4.bottomOffset:		4",
"*hr.bottomOffset:		4",
"*XmComboBox.marginWidth:	0",
"*XmComboBox.marginHeight:	0",
"*demo.marginWidth:		30",
"*demo.marginHeight:		4",
"*man.marginWidth:		10",
"*man.marginHeight:		4",
"*down.leftOffset:		40",
"*down.marginWidth:		4",
"*down.marginHeight:		4",
"*up.marginWidth:		4",
"*up.marginHeight:		4",
"*frame.traversalOn:		False",
"*list.automaticSelection:	True",
"*list.visibleItemCount:		20",
"*doc.columns:			60",
"*combo.columns:			11",
"*demoTab.labelString:		Graphics Demos",
"*optionsTab.labelString:	Screensaver Options",
"*down.labelString:		\\\\/ ",
"*up.labelString:		/\\\\ ",
"*frameLabel.labelString:	",
"*cmdLabel.labelString:		Command Line:",
"*cmdLabel.alignment:		ALIGNMENT_BEGINNING",
"*enabled.labelString:		Enabled",
"*visLabel.labelString:		Visual:",
"*visLabel.alignment:		ALIGNMENT_END",
"*visLabel.leftOffset:		20",
"*demo.labelString:		Demo",
"*man.labelString:		Documentation...",
"*done.labelString:		Quit",
"*preferencesLabel.labelString:	XScreenSaver Parameters",
"*timeoutLabel.labelString:	Saver Timeout",
"*cycleLabel.labelString:	Cycle Timeout",
"*fadeSecondsLabel.labelString:	Fade Duration",
"*fadeTicksLabel.labelString:	Fade Ticks",
"*lockLabel.labelString:		Lock Timeout",
"*passwdLabel.labelString:	Password Timeout",
"*preferencesForm*XmTextField.columns:	8",
"*verboseToggle.labelString:	Verbose",
"*cmapToggle.labelString:	Install Colormap",
"*fadeToggle.labelString:	Fade Colormap",
"*unfadeToggle.labelString:	Unfade Colormap",
"*lockToggle.labelString:	Require Password",
"*OK.marginWidth:		30",
"*OK.marginHeight:		4",
"*OK.leftOffset:			10",
"*OK.bottomOffset:		10",
"*Cancel.marginWidth:		30",
"*Cancel.marginHeight:		4",
"*Cancel.rightOffset:		10",
"*Cancel.bottomOffset:		10",
